package ru.dbotthepony.mc.otm.client.render

import com.mojang.blaze3d.systems.RenderSystem
import com.mojang.blaze3d.vertex.DefaultVertexFormat
import com.mojang.blaze3d.vertex.VertexFormat
import net.minecraft.client.renderer.GameRenderer
import net.minecraft.client.renderer.RenderStateShard
import net.minecraft.client.renderer.RenderStateShard.TextureStateShard
import net.minecraft.client.renderer.RenderType
import net.minecraft.client.renderer.texture.TextureAtlasSprite
import net.minecraft.resources.ResourceLocation
import net.minecraft.util.Mth
import org.lwjgl.opengl.GL11.GL_ALWAYS
import ru.dbotthepony.mc.otm.isClient
import java.util.stream.Stream

class AtlasSkinElement(
	val location: ResourceLocation,
	override val width: Float,
	override val height: Float,
	val spriteWidth: Float = Mth.smallestEncompassingPowerOfTwo(width.toInt()).toFloat(),
	val spriteHeight: Float = Mth.smallestEncompassingPowerOfTwo(height.toInt()).toFloat(),
	override val winding: UVWindingOrder = UVWindingOrder.NORMAL,
) : AbstractSkinElement() {
	init {
		synchronized(keys) {
			if (keys.add(location)) {
				if (isWidgetAtlasAvailable) {
					queueRebuild()
				}
			}
		}
	}

	private var changeset = -1
	private var _textureAtlasSprite: TextureAtlasSprite? = null

	private val textureAtlasSpriteImpl: TextureAtlasSprite get(): TextureAtlasSprite {
		check(isClient) { "Invalid realm" }
		val _textureAtlasSprite = _textureAtlasSprite

		if (_textureAtlasSprite == null || _textureAtlasSprite.name.let { it.namespace == "minecraft" && it.path == "missingno" } || changeset != WidgetAtlasHolder.INSTANCE.changeset || WidgetAtlasHolder.INSTANCE.speculateIfMinecraftEngineIsDumDum()) {
			val get = WidgetAtlasHolder.INSTANCE.getSprite(location)
			this._textureAtlasSprite = get
			changeset = WidgetAtlasHolder.INSTANCE.changeset
			return get
		}

		return _textureAtlasSprite
	}

	val textureAtlasSprite: TextureAtlasSprite get(): TextureAtlasSprite {
		check(isWidgetAtlasAvailable) { "Atlas is not available; either we are not a client, or we are running datagen" }
		return textureAtlasSpriteImpl
	}

	override val u0 get() = textureAtlasSprite.u0
	override val v0 get() = textureAtlasSprite.v0
	override val u1 get() = textureAtlasSprite.u1 - (textureAtlasSprite.u1 - textureAtlasSprite.u0) * (1f - width / spriteWidth)
	override val v1 get() = textureAtlasSprite.v1 - (textureAtlasSprite.v1 - textureAtlasSprite.v0) * (1f - height / spriteHeight)
	override val texture: ResourceLocation get() = WidgetAtlasHolder.LOCATION

	override fun equals(other: Any?): Boolean {
		if (other is AtlasSkinElement)
			return location == other.location

		return super.equals(other)
	}

	override fun hashCode(): Int {
		return WidgetAtlasHolder.INSTANCE.hashCode() xor location.hashCode()
	}

	override fun toString(): String {
		return "AtlasSkinElement[$location]"
	}

	override val type: SkinElementType
		get() = SkinElementType.ATLAS

	companion object {
		private val keys = HashSet<ResourceLocation>()
		val keysStream: Stream<ResourceLocation> get() = keys.stream()

		private fun queueRebuild() {
			WidgetAtlasHolder.INSTANCE.queueRebuild()
		}

		val renderTypeNoDepth by lazy {
			val builder = RenderType.CompositeState.builder()

			builder.setTextureState(TextureStateShard(WidgetAtlasHolder.LOCATION, false, false))
			builder.setShaderState(RenderStateShard.ShaderStateShard(GameRenderer::getPositionColorTexShader))
			builder.setDepthTestState(RenderStateShard.DepthTestStateShard("always", GL_ALWAYS))
			builder.setTransparencyState(RenderStateShard.TransparencyStateShard("normal_blend", {
				RenderSystem.enableBlend()
				RenderSystem.defaultBlendFunc()
			}, {
				RenderSystem.disableBlend()
			}))

			@Suppress("INACCESSIBLE_TYPE")
			RenderType.create("otm_gui_element_no_depth",
				DefaultVertexFormat.POSITION_COLOR_TEX,
				VertexFormat.Mode.QUADS,
				2048,
				false,
				false,
				builder.createCompositeState(false)) as RenderType
		}

		val renderTypeDepth by lazy {
			val builder = RenderType.CompositeState.builder()

			builder.setTextureState(TextureStateShard(WidgetAtlasHolder.LOCATION, false, false))
			builder.setShaderState(RenderStateShard.ShaderStateShard(GameRenderer::getPositionColorTexShader))
			builder.setTransparencyState(RenderStateShard.TransparencyStateShard("normal_blend", {
				RenderSystem.enableBlend()
				RenderSystem.defaultBlendFunc()
			}, {
				RenderSystem.disableBlend()
			}))

			@Suppress("INACCESSIBLE_TYPE")
			RenderType.create("otm_gui_element_depth",
				DefaultVertexFormat.POSITION_COLOR_TEX,
				VertexFormat.Mode.QUADS,
				2048,
				false,
				false,
				builder.createCompositeState(false)) as RenderType
		}

		val renderTypeWorld by lazy {
			val builder = RenderType.CompositeState.builder()

			builder.setTextureState(TextureStateShard(WidgetAtlasHolder.LOCATION, false, false))
			builder.setShaderState(RenderStateShard.ShaderStateShard(GameRenderer::getPositionColorTexShader))
			builder.setTransparencyState(RenderStateShard.TransparencyStateShard("normal_blend", {
				RenderSystem.enableBlend()
				RenderSystem.defaultBlendFunc()
			}, {
				RenderSystem.disableBlend()
			}))

			@Suppress("INACCESSIBLE_TYPE")
			RenderType.create("otm_gui_element_world",
				DefaultVertexFormat.POSITION_COLOR_TEX,
				VertexFormat.Mode.QUADS,
				8192,
				false,
				false,
				builder.createCompositeState(false)) as RenderType
		}
	}
}
